/*  Calculate Greenspan/Barnes/Hut accelerations.			*/
/*  Source file:  accel3.cs		Prototypes:  accel3.hs		*/
/*  Tree should have been constructed and filled as in tree.cs, fill.cs	*/
/*  Uses declarations in gbh.hs.					*/
/*  This version:							*/
/*   *	starts each particle at it's own node and continues		*/
/*	until particles return to their nodes.				*/
/*   *  buffers 3-vectors for gets					*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <cscomm.h>
#include "gbh.hs"
#include "acdiag.hs"


/*  Calculate accelerations according to error parameters,	*/
/*  theta, beta, tree data Xp0, Xp1, Tm, Th.			*/
/*  Call from Particles with all processors active.		*/
/*  Returns number of cycles of tree traversal
/*  Leaves accelerations in Ac,  statistics on calculation in	*/
/*	nodex - nodes examined and rejected			*/
/*	nodea - nodes examined and accepted			*/
/*	cactv - total active vp cycles
/*  Doubles allow for very large numbers.  Ints are sufficient	*/
/*  for counting in individual vp's.				*/
/*  The (usually N) times that a node examines its own leaf	*/
/*  are ignored in both counts.					*/

#define Tm (*pTm)
#define Th (*pTh)

int Accel3(float:Tree Xp0[3], float:Tree Xp1[3], float:Tree *pTm,
           float:Tree *pTh, float theta, float beta,
	   double *nodex, double *nodea)	/* For analysis */
 { float:Particles D[3],	/* D0 + D1		*/
	   B0[3], B1[3],	/* Comm. buffers	*/
       r0, r1, rd, v, w;	/* Distances, working	*/
   int:Particles  ar,		/* 1: accept, 2: reject	*/
		  nd,		/* Tree node		*/
            prx, pra;		/* Counters (analysis)	*/
   CMC_sendaddr_t:Particles sn;	/* Address of [nd]Tree	*/
   int acrj, und, unx, u, cyc;

   everywhere
   { prx = pra  = ar = 0;  cyc = 0;	/* Cycle counter */
     for ( u = 0; u < 3; u++ ) Ac[u] = 0.0f;	/* Clear Ac      */
     with (Tree) everywhere
       where ( Tnext < 0 ) Tnext = 1;	/* Traversals "wrap around" */
     nd = [cell]Tnext;
	/* Loop to compute acceleration, particles act independently	 */
	/* Each particle is ready to examine node nd (== cell if done)	 */
	/* Not all particles are finished.				 */
	do
	 { cyc++;  r0 = r1 = rd = 0.0f;		/* Everywhere	*/
	   /* printf("Cycle %d\n", cyc);	    /*** Testing ***/
	   sn = make_send_address(Tree, nd);
	   where ( nd != cell )		/* Particle is still active	*/
	    { get(B0, &sn, Xp0, CMC_collisions, pvsz);
	      get(B1, &sn, Xp1, CMC_collisions, pvsz);
	    }
	   for ( u = 0; u < 3; u++ )
	    { v = B0[u] - X0[u];  r0 += v*v;
	      w = B1[u] - X1[u];  D[u] = v += w;
	      r1 += w*w;  rd += v*v;
	    }
	   r0 = sqrt(r0);  r1 = sqrt(r1);  rd = sqrt(rd);  w = r0 + r1;
   	   where ( nd != cell )		/* Particle is still active	*/
	    { where ( get(sn, pTh, CMC_collisions) <= theta*(r0 <? r1)
		     & (rd >= beta*w) 				      )
	       { w = get(sn, pTm, CMC_collisions)/(r0*r1*w);
	         pra++;	         
	         for ( u = 0; u < 3; u++ )  Ac[u] += w*D[u];
	         nd = get(sn, pTnext, CMC_collisions);
	       }
	      else	/* Rejected nd */
	       { nd = get(sn, pChild0, CMC_collisions);  prx++; }
	    }	/* end where ( nd > 0 )	*/
	 }
	while ( (|= (nd != cell)) & cyc <= Nodes );	/* Safety */
     with (Tree) everywhere
       where ( Tnext == 1 ) Tnext = -1;	/* Restore tree */
     /* Report statistics */
     *nodex = (+= (double:current)prx);
     *nodea = (+= (double:current)pra);
     return (cyc);
   }
 }
